from pyduinocli.commands.base import CommandBase
from pyduinocli.constants import commands
from pyduinocli.constants import flags


class ConfigCommand(CommandBase):

    def __init__(self, base_args):
        CommandBase.__init__(self, base_args)
        self._base_args.append(commands.CONFIG)

    def dump(self):
        return self._exec([commands.DUMP])

    def init(self, save_as=None):
        args = [commands.INIT]
        if save_as:
            args.extend([flags.SAVE_AS, CommandBase._strip_arg(save_as)])
        return self._exec(args)



from pyduinocli.commands.base import CommandBase
from pyduinocli.constants import commands


class VersionCommand(CommandBase):

    def __init__(self, base_args):
        CommandBase.__init__(self, base_args)
        self._base_args.append(commands.VERSION)

    def __call__(self):
        return self._exec([])



from pyduinocli.commands.base import CommandBase
from pyduinocli.constants import commands
from pyduinocli.constants import flags


class CompileCommand(CommandBase):

    def __init__(self, base_args):
        CommandBase.__init__(self, base_args)
        self._base_args.append(commands.COMPILE)

    def __call__(self,
                 sketch, build_cache_path=None, build_path=None, build_properties=None, fqbn=None, output=None,
                 port=None, preprocess=None, show_properties=None, upload=None, verify=None, vid_pid=None,
                 warnings=None):
        args = []
        if build_cache_path:
            args.extend([flags.BUILD_CACHE_PATH, CommandBase._strip_arg(build_cache_path)])
        if build_path:
            args.extend([flags.BUILD_PATH, CommandBase._strip_arg(build_path)])
        if build_properties:
            args.extend([flags.BUILD_PROPERTIES, CommandBase._strip_arg(build_properties)])
        if fqbn:
            args.extend([flags.FQBN, CommandBase._strip_arg(fqbn)])
        if output:
            args.extend([flags.OUTPUT, CommandBase._strip_arg(output)])
        if port:
            args.extend([flags.PORT, CommandBase._strip_arg(port)])
        if preprocess is True:
            args.append(flags.PREPROCESS)
        if show_properties is True:
            args.append(flags.SHOW_PROPERTIES)
        if upload is True:
            args.append(flags.UPLOAD)
        if verify is True:
            args.append(flags.VERIFY)
        if vid_pid:
            args.extend([flags.VID_PID, CommandBase._strip_arg(vid_pid)])
        if warnings:
            args.extend([flags.WARNINGS, CommandBase._strip_arg(warnings)])
        args.append(CommandBase._strip_arg(sketch))
        return self._exec(args)



from pyduinocli.commands.base import CommandBase
from pyduinocli.constants import commands
from pyduinocli.constants import flags


class CoreCommand(CommandBase):

    def __init__(self, base_args):
        CommandBase.__init__(self, base_args)
        self._base_args.append(commands.CORE)
        
    def download(self, downloads):
        args = [commands.DOWNLOAD]
        args.extend(CommandBase._strip_args(downloads))
        return self._exec(args)

    def install(self, installs):
        args = [commands.INSTALL]
        args.extend(CommandBase._strip_args(installs))
        return self._exec(args)

    def list(self, updatable=None):
        args = [commands.LIST]
        if updatable is True:
            args.append(flags.UPDATABLE)
        return self._exec(args)

    def search(self, keywords=None, all=None):
        args = [commands.SEARCH]
        if keywords is None:
            keywords = []
        if keywords:
            args.extend(CommandBase._strip_args(keywords))
        if all is True:
            args.append(flags.ALL)
        return self._exec(args)

    def uninstall(self, installs):
        args = [commands.UNINSTALL]
        args.extend(CommandBase._strip_args(installs))
        return self._exec(args)

    def update_index(self):
        return self._exec([commands.UPDATE_INDEX])

    def upgrade(self, upgrades=None):
        if not upgrades:
            upgrades = []
        args = [commands.UPGRADE]
        args.extend(CommandBase._strip_args(upgrades))
        return self._exec(args)



from pyduinocli.commands.base import CommandBase
from pyduinocli.constants import commands
from pyduinocli.constants import messages
from pyduinocli.constants import flags
from pyduinocli.errors import ArduinoError


class BoardCommand(CommandBase):

    def __init__(self, base_args):
        CommandBase.__init__(self, base_args)
        self._base_args.append(commands.BOARD)

    def attach(self, port=None, fqbn=None, sketch_path=None, timeout=None):
        args = [commands.ATTACH]
        if port and fqbn:
            raise ArduinoError(messages.ERROR_PORT_FQBN_SET)
        if not port and not fqbn:
            raise ArduinoError(messages.ERROR_PORT_FQBN_NOT_SET)
        if port:
            args.append(CommandBase._strip_arg(port))
        if fqbn:
            args.append(CommandBase._strip_arg(fqbn))
        if sketch_path:
            args.append(CommandBase._strip_arg(sketch_path))
        if timeout:
            args.extend([flags.TIMEOUT, CommandBase._strip_arg(timeout)])
        return self._exec(args)

    def details(self, fqbn):
        args = [commands.DETAILS, CommandBase._strip_arg(fqbn)]
        return self._exec(args)

    def list(self, timeout=None):
        args = [commands.LIST]
        if timeout:
            args.extend([flags.TIMEOUT, CommandBase._strip_arg(timeout)])
        return self._exec(args)

    def listall(self, boardname=None):
        args = [commands.LISTALL]
        if boardname:
            args.append(CommandBase._strip_arg(boardname))
        return self._exec(args)



from pyduinocli.commands.base import CommandBase
from pyduinocli.constants import commands
from pyduinocli.constants import flags


class DaemonCommand(CommandBase):

    def __init__(self, base_args):
        CommandBase.__init__(self, base_args)
        self._base_args.append(commands.DAEMON)

    def __call__(self, daemonize=None, port=None):
        args = []
        if daemonize is True:
            args.append(flags.DAEMONIZE)
        if port:
            args.extend([flags.PORT, CommandBase._strip_arg(str(port))])
        return self._exec(args)



ADDITIONAL_URLS = '--additional-urls'
CONFIG_FILE = '--config-file'
FORMAT = '--format'
LOG_FILE = '--log-file'
LOG_FORMAT = '--log-format'
LOG_LEVEL = '--log-level'
TIMEOUT = '--timeout'
BUILD_CACHE_PATH = '--build-cache-path'
BUILD_PATH = '--build-path'
BUILD_PROPERTIES = '--build-properties'
FQBN = '--fqbn'
OUTPUT = '--output'
UPLOAD = '--upload'
PREPROCESS = '--preprocess'
SHOW_PROPERTIES = '--show-properties'
VID_PID = '--vid-pid'
WARNINGS = '--warnings'
SAVE_AS = '--save-as'
VERIFY = '--verify'
INPUT = '--input'
PORT = '--port'
UPDATABLE = '--updatable'
NAMES = '--names'
ALL = '--all'
DAEMONIZE = '--daemonize'



from pyduinocli.commands.base import CommandBase
from pyduinocli.commands.board import BoardCommand
from pyduinocli.commands.compile import CompileCommand
from pyduinocli.commands.config import ConfigCommand
from pyduinocli.commands.core import CoreCommand
from pyduinocli.commands.daemon import DaemonCommand
from pyduinocli.commands.lib import LibCommand
from pyduinocli.commands.sketch import SketchCommand
from pyduinocli.commands.upload import UploadCommand
from pyduinocli.commands.version import VersionCommand
from pyduinocli.constants import flags


class ArduinoCliCommand(CommandBase):

    __FORMAT_JSON = 'json'

    def __init__(self, cli_path="arduino-cli", config_file=None, additional_urls=None, log_file=None, log_format=None,
                 log_level=None):
        base_args = [cli_path, flags.FORMAT, ArduinoCliCommand.__FORMAT_JSON]
        if config_file:
            base_args.extend([flags.CONFIG_FILE, CommandBase._strip_arg(config_file)])
        if additional_urls:
            base_args.extend([flags.ADDITIONAL_URLS, ",".join(CommandBase._strip_args(additional_urls))])
        if log_file:
            base_args.extend([flags.LOG_FILE, CommandBase._strip_arg(log_file)])
        if log_format:
            base_args.extend([flags.LOG_FORMAT, CommandBase._strip_arg(log_format)])
        if log_level:
            base_args.extend([flags.LOG_LEVEL, CommandBase._strip_arg(log_level)])
        CommandBase.__init__(self, base_args)
        self.board = BoardCommand(self._base_args)
        self.compile = CompileCommand(self._base_args)
        self.config = ConfigCommand(self._base_args)
        self.core = CoreCommand(self._base_args)
        self.daemon = DaemonCommand(self._base_args)
        self.lib = LibCommand(self._base_args)
        self.sketch = SketchCommand(self._base_args)
        self.upload = UploadCommand(self._base_args)
        self.version = VersionCommand(self._base_args)



import re
import json
from subprocess import Popen, PIPE
from pyduinocli.errors import ArduinoError
from pyduinocli.constants import messages


class CommandBase:
    __ERROR_MESSAGE = "Message"
    __ERROR_CAUSE = "Cause"

    __REGEX_JSON = r'(\{|\[)'

    def __init__(self, base_args):
        self._base_args = list(base_args)

    @staticmethod
    def _strip_arg(arg):
        return arg.lstrip("-")

    @staticmethod
    def _strip_args(args):
        out = list()
        for arg in args:
            out.append(CommandBase._strip_arg(arg))
        return out

    @staticmethod
    def __parse_output(data):
        pattern = re.compile(CommandBase.__REGEX_JSON)
        for match in pattern.finditer(data):
            candidate = data[match.start():]
            try:
                return json.loads(candidate)
            except ValueError:
                continue
        return data

    def _exec(self, args):
        command = list(self._base_args)
        command.extend(args)
        try:
            p = Popen(command, stdout=PIPE, stderr=PIPE, universal_newlines=True)
            stdout, stderr = p.communicate()
            stdout = stdout.strip()
            stderr = stderr.strip()
            decoded_out = self.__parse_output(stdout)
            if p.returncode != 0:
                if type(decoded_out) is dict:
                    raise ArduinoError(decoded_out[CommandBase.__ERROR_MESSAGE],
                                       decoded_out[CommandBase.__ERROR_CAUSE],
                                       stderr)
                raise ArduinoError(decoded_out,
                                   None,
                                   stderr)
            return decoded_out
        except OSError:
            raise ArduinoError(messages.ERROR_OSERROR % self._base_args[0])



from pyduinocli.commands.base import CommandBase
from pyduinocli.constants import commands
from pyduinocli.constants import flags


class UploadCommand(CommandBase):

    def __init__(self, base_args):
        CommandBase.__init__(self, base_args)
        self._base_args.append(commands.UPLOAD)

    def __call__(self, sketch=None, fqbn=None, input=None, port=None, verify=None):
        args = []
        if fqbn:
            args.extend([flags.FQBN, CommandBase._strip_arg(fqbn)])
        if input:
            args.extend([flags.INPUT, CommandBase._strip_arg(input)])
        if port:
            args.extend([flags.PORT, CommandBase._strip_arg(port)])
        if verify is True:
            args.append(flags.VERIFY)
        if sketch:
            args.append(CommandBase._strip_arg(sketch))
        return self._exec(args)



from pyduinocli.commands.base import CommandBase
from pyduinocli.constants import commands


class SketchCommand(CommandBase):

    def __init__(self, base_args):
        CommandBase.__init__(self, base_args)
        self._base_args.append(commands.SKETCH)

    def new(self, name):
        return self._exec([commands.NEW, CommandBase._strip_arg(name)])
